print("import: 'google.api_core.client_info'")
import google.api_core.client_info

print("import: 'google.cloud.client'")
import google.cloud.client

print("import: 'google.cloud.environment_vars'")
import google.cloud.environment_vars

print("import: 'google.cloud.exceptions'")
import google.cloud.exceptions

print("import: 'google.cloud.operation'")
import google.cloud.operation

